// Initial Developer: 2002 Takeshi Morimoto <morimoto@takopen.cs.uec.ac.jp> 
// Author: Zijian Ren
// Version :1.20
// Date: Sep 17, 2003
// Comments: Avoid Buring Building
//		adjust rescue and load commands



package phoenix;

import java.net.*;
import java.util.*;
import yab.agent.*;
import yab.agent.object.*;

public class AmbulanceTeamAgent extends AbstractAmbulanceTeamAgent {
    private HashSet unvisitedBuildings=new HashSet();
    private HashSet buriedBuildings = new HashSet();
    private HashSet urgentBuildings = new HashSet();
    private HashSet blockedRoads = new HashSet();
    
    private HashSet localHumans = new HashSet();
        
    private int status;
    private static final int CARRY=1, RESCUE=2, SEARCH=3, URGENT=4;
    private int hearMessages;
    private int civilMessages =0, prevCivilMessages=0;
    private int prevDegree = 0;
    
    private Humanoid targetHuman, loadedHuman;
    

    public AmbulanceTeamAgent(InetAddress address, int port) {
        super(address, port);

	status = SEARCH;
	
	unvisitedBuildings.addAll(world.buildings);
	unvisitedBuildings.removeAll(world.refuges);
	
	/*		
	for (Iterator it=world.buildings.iterator(); it.hasNext(); ) {
		Building b=(Building)it.next();
		if (!(b instanceof Refuge)) 
			unvisitedBuildings.add(b);
	}		
	*/

    }

    protected MotionlessObject mlpos() { return self().motionlessPosition(); }

    protected static final Property
        DAMAGE_PRP     = Property.get("Humanoid", "damage"),
        BURIEDNESS_PRP = Property.get("Humanoid", "buriedness"),
        /*add health point*/
	HP_PRP     = Property.get("Humanoid", "hp"),

	POSITION_PRP   = Property.get("MovingObject", "position"),
        FIERYNESS_PRP  = Property.get("Building", "fieryness");

    protected static final Condition
        IS_NOT_BURNING_CND = FIERYNESS_PRP.lt(1).or(FIERYNESS_PRP.gt(3));

	
    protected void prepareForAct() {
        super.prepareForAct();
        if (mlpos() instanceof Building)
            unvisitedBuildings.remove(mlpos());
    }
    

    protected void act() throws ActionCommandException {
     
  	hearMessages = 0;
	
	prevCivilMessages = civilMessages;
	civilMessages = 0;
    
    	checkSelf();
	
	//Methods in class Humanoid
	if (checkTrafficJam()) {
		moveToBuilding();
	}
	
        if (isLoadingInjured()) {
		status = CARRY;
        	carry();
	}

	if (status == URGENT) {
		rescue_load(targetHuman);
	}
	
	//check current building is burning
	if (reportBuringBuilding(mlpos()))
		wander();

	//System.out.println("Time "+time()+urgentBuildings);
				
	if (status == RESCUE) {
		//if there are urgent buildings and current position is not there
		if (!urgentBuildings.isEmpty() && !urgentBuildings.contains(mlpos()) ) {
			//flag 1 indicates urgent
			move_rescue(urgentBuildings);
		} else {
			rescue_load(targetHuman);
		}		
	}
	
	if (status == SEARCH) {
		findInjured();
		wander();
	}
		
    }
   
 
    
    protected void checkSelf() throws ActionCommandException {
    	if (self().hp() == 0)
		rest();
	
	if (mlpos() instanceof Road) {
		int degree = 1;

		
		if (prevDegree > 0)
			degree = 5;
			
		Road r = (Road)mlpos();
		if ( r.passableLines() == 0  ) {
			if (!blockedRoads.contains(r)) {
				blockedRoads.add(r);
				//System.out.println(r+" block "+r.block()+" repairCost "+r.repairCost()+" tohead "+r.linesToHead() +
				//" totail "+r.linesToTail()+" passable "+r.passableLinesToHead()+" "+r.passableLinesToTail());
				tell("time " + time() + " clear " + r.id + " degree "+degree);
				if (degree == 5)
					prevDegree = 0;
				else
					prevDegree = degree;	
			}	
		}
	}
	
	if (time()<20) {
	for(Iterator it=world.buildings.iterator(); it.hasNext(); ) {
		Building b = (Building)it.next();
		
		if (b.isBurning() && !buriedBuildings.contains(b)) {
				//System.out.println("Time "+time()+" "+self()+" tell: extinguish "+b.id);
				buriedBuildings.add(b);
				tell("time "+time()+" extinguish "+b.id);
				break;
		}
	} //end for
 	}
	
	if (self().buriedness()>0 || self().damage()>0) {
		System.out.println("!!!Time "+time()+" "+self()+" buried "+self().buriedness()+" hp "+self().hp());
		if (self().buriedness()>0)
			rescue(self());
			
		move(world.refuges);
	}
    }
    
    protected void carry() throws ActionCommandException {
        if (mlpos() instanceof Refuge) {
            System.out.println("Time "+time()+" "+self()+" "+mlpos()+" unload "+loadingInjured());
	    status = SEARCH;
	    unload();
        } else {
		System.out.println("Time "+time()+" "+self()+" "+mlpos()+" move to refuge!");
		status = CARRY;
       		move(world.refuges);
	}	
    }

    protected void rescue_load(Humanoid target) throws ActionCommandException{
    	status = RESCUE;
    	//check whether h and self are the same position or the h information is obsolete
	if (target.motionlessPosition() != mlpos() || target.time() != time()) {
		System.out.println("Time "+time()+" "+self()+" cannot load "+target+" SEARCH again!!!");
		status = SEARCH;
		return;
	}	   	
	
	if (target.buriedness() == 0) {
		//for multiple ambulances, min id loads the person
		List ambulance = POSITION_PRP.eq(mlpos()).extract(world.ambulanceTeams);
		ambulance.remove(self());
		
		//judge load or not
		boolean bLoad = true;
		for(Iterator it=ambulance.iterator(); it.hasNext(); ) {
			if ( ((Humanoid)it.next()).id < self().id )
				bLoad = false;
		}	 
		
		//add for firebrigade situation
		if (target instanceof FireBrigade)
			bLoad = false;
			
		if (bLoad) {
			if (localHumans.isEmpty()) {
				//the building is empty with rescued persons
				System.out.println("!!!Time "+time()+" "+self()+" Remove "+mlpos()+" and tell a message");
				buriedBuildings.remove(mlpos());
				urgentBuildings.remove(mlpos());
				unvisitedBuildings.remove(mlpos());
				//??? multiple remove command???
				tell("time "+time()+" remove "+mlpos().id);
			}
				
			System.out.println("!!!Time "+time()+" "+self()+" "+mlpos()+
				" load "+target+" hp "+target.hp()+"  buriedness "+target.buriedness()+" damage "+target.damage());
			status = CARRY;
			load(target);
		} else {
			//System.out.println("Time "+time()+" "+self()+" "+mlpos()+" cannot load "+target);
			if (localHumans.isEmpty()) {
				//System.out.println("!!!Time "+time()+" "+self()+" empty "+mlpos());
				buriedBuildings.remove(mlpos());
				urgentBuildings.remove(mlpos());
				
				tell("time "+time()+" empty "+mlpos().id);
				wander();
			} else {
				//find the other person in localHuamns
				//targetHuman = (Humanoid)BURIEDNESS_PRP.min(localHumans);
				targetHuman = (Humanoid)DAMAGE_PRP.max(localHumans);
				localHumans.remove(targetHuman);
				System.out.println("Time "+time()+" "+self()+" "+mlpos()+" another target "+targetHuman+" hp "+targetHuman.hp()+"  buriedness "+targetHuman.buriedness()+" damage "+targetHuman.damage());
				
				status = RESCUE;
				rescue(targetHuman);
			}
		}		
    	} else {
		rescue(target);
	} // end buriedness 0	
    }
	
    private void printHumanoid(Set humans) {
    	System.out.print("Time "+time()+" Ambulance "+self().id+": ");
    	for(Iterator it=humans.iterator(); it.hasNext(); ) {
		Humanoid h= (Humanoid)it.next();
		System.out.print(h+" hp "+h.hp()+"  buriedness "+h.buriedness()+" damage "+h.damage()+"; ");
	}
    	System.out.println();
    }
 
       protected boolean reportBuringBuilding(MotionlessObject ml) {
	int fieryness = (ml instanceof Building)? ((Building)ml).fieryness(): 0;
	if (fieryness >=1  && fieryness <= 3) {
		System.out.println("WARNING! Time "+time()+" "+self()+" hp "+self().hp()+" status "+status+" avoid and report "+ml+" fieryness "+fieryness);
		//tell("time "+time()+" extinguish "+mlpos().id);
		return true;
	}
	
	return false;
    }   
    
    protected void move_rescue(Collection buildings) throws ActionCommandException{
    	
    	if ( buildings.contains(mlpos()) ) {
		findInjured();	
	} else {
		move(buildings);
	}				
    }
    
    protected void findInjured() throws ActionCommandException{
   	localHumans.clear();
	FireBrigade f = null; 
	
	if (mlpos() instanceof Refuge)
		wander();
		
	//check: damage>0, hp>0, hp>damage*buriedness, and NOT in refuge
	for(Iterator it=world.humanoids.iterator(); it.hasNext(); ) {
		Humanoid h = (Humanoid)it.next();
		
		if (h.hp() > 0 && h.damage() >0 && h.time() == time()) {
			MotionlessObject ml = h.motionlessPosition();
			//tell the building
			if (ml instanceof Building && !(ml instanceof Refuge)) {
				//System.out.println("Time "+time()+" "+self()+" tell "+ml);
				unvisitedBuildings.remove(ml);
				buriedBuildings.add(ml);
				tell("time "+time()+" locate "+ml.id);
				

				if (ml == mlpos()) {
					if (h instanceof FireBrigade)
						f = (FireBrigade)h;
					localHumans.add(h);
				}	
			}	
		}
	
	} // end for
	

	//printHumanoid(localHumans);
	//printHumanoid(aroundHumans);
	
	if (localHumans.isEmpty()) {
		urgentBuildings.remove(mlpos());
		buriedBuildings.remove(mlpos());
		unvisitedBuildings.remove(mlpos());
		return;
	}
							
	//If there are humans in the building and building is not burning
	if (!reportBuringBuilding(mlpos())) {
		//get target and other humans in the same place
		//targetHuman = (Humanoid)BURIEDNESS_PRP.min(localHumans);
		if (f != null) {
			status = URGENT;
			targetHuman = f;
		} else	{
			status = RESCUE;
			targetHuman = (Humanoid)DAMAGE_PRP.max(localHumans);
		}
		
		localHumans.remove(targetHuman);
		
		//SHOW LATER	
		//System.out.println("Time "+time()+" "+self()+" "+mlpos()+" target "+targetHuman+" hp "+targetHuman.hp()+"  buriedness "+targetHuman.buriedness()+" damage "+targetHuman.damage());
		rescue_load(targetHuman);
	}	
	
    }

    protected void wander() throws ActionCommandException{
    //priority: humanBuildings, unvisitedBuildings
    
    	status = SEARCH;
    	//System.out.println("Time "+time()+" "+self()+" "+mlpos()+" wander!!!");

	if (!urgentBuildings.isEmpty()) {
		move_rescue(urgentBuildings);
	} else {
		List buried = IS_NOT_BURNING_CND.extract(buriedBuildings);
		if (!buried.isEmpty()) {
			//System.out.println("Time "+time()+" "+self()+" unburnt humanBuildings "+bldgs);
			move_rescue(buried);
		}
	}	
	
	MotionlessObject ml = mlpos();
	//use previous civil messages to determin serach areas
	HashSet nearBuildings = new HashSet();
	for(Iterator it=world.buildings.iterator(); it.hasNext(); ) {
		Building b = (Building)it.next();
		//say distance 30*1000, less error 100
		if (b.distance(ml) < (30 * 1000 - 100) )
			nearBuildings.add(b);
	}		
		
	if (prevCivilMessages == 0) {
		unvisitedBuildings.removeAll(nearBuildings);
		List unvisited = IS_NOT_BURNING_CND.extract(unvisitedBuildings);
		if (!unvisited.isEmpty())
			move(unvisited);
	} else {
		HashSet targets = new HashSet(); 
		for(Iterator it=nearBuildings.iterator(); it.hasNext(); ) {
			Building b = (Building)it.next();
			if (unvisitedBuildings.contains(b) && !b.isBurning())
				targets.add(b);
		}
		
		if (!targets.isEmpty())
			move(targets);		
	}	
	
	System.out.println("Time "+time()+" "+self()+" NO PLACE TO GO");		
    }
    
    
    protected void hear(RealObject sender, String message) {
    
       //System.out.println("Time "+time()+" receiver "+self()+" sender "+sender+" "+message);
	if (hearMessages >= hearingLimit())
		return;
		
	//only count civilian say help messages
	if (sender instanceof Civilian)
		civilMessages++;
		
/*       
    	if (sender instanceof AmbulanceTeam) {
		StringTokenizer st = new StringTokenizer(message);
		st.nextToken();
        	st.nextToken();
		
		String command = st.nextToken();
		
		if (command.equals("locate")) {
			while (st.hasMoreTokens()) {
				RealObject target = world.get(Integer.parseInt(st.nextToken()));
            			buriedBuildings.add(target);
			} 
		} else if (command.equals("empty")) {
			while (st.hasMoreTokens()) {
				RealObject target = world.get(Integer.parseInt(st.nextToken()));
            			buriedBuildings.remove(target);
				urgentBuildings.remove(target);
			}
		}		
	} // end if
*/
	
	//Buried information from ambulcance center is ugent because it comes from FB or PF	
	if (sender instanceof AmbulanceCenter) {
	//System.out.println("Time "+time()+" receiver "+self()+" sender "+sender+" "+message);
		hearMessages++;
		
		StringTokenizer st = new StringTokenizer(message);
		st.nextToken(); //String "Time"
        	st.nextToken(); //value of time
		
		String command = st.nextToken();
		
		if (command.equals("locate")) {
			int degree = 1;
			//locate degree 5 x1 x2 ... degree 1 y1 y2  
			while(st.hasMoreTokens()) {
				String s = st.nextToken();
				if (s.equals("degree")) {
					degree = Integer.parseInt(st.nextToken());
				} else {
					RealObject target = world.get(Integer.parseInt(s));
					//Temporary use two degree:urgent and buried
					if (degree == 5) {
						urgentBuildings.add(target);
					} else {
						buriedBuildings.add(target);
					}
							
					unvisitedBuildings.remove(target);

				}
			} //end while
		} else if (command.equals("empty")) {
			while (st.hasMoreTokens()) {
				RealObject target = world.get(Integer.parseInt(st.nextToken()));
            			buriedBuildings.remove(target);
				urgentBuildings.remove(target);
				unvisitedBuildings.remove(target);
			}

		} // end if command 
    	} // end if sender
    } // end of hear()		
}
