// Initial Developer: 2002 Takeshi Morimoto <morimoto@takopen.cs.uec.ac.jp> 
// Author: Zijian Ren
// Version :1.30
// Date: July 16, 2003
// Comments:	Add water refill
//		Add refill water amount
//		Add timestamp in a message
//
// Fire Messages: time 10 id 1099 (degree d)
// Main ideas:
// Check itself, check fires
// Supply Water
// Extinguish status: if target is put out, find a new fire, else continue extinguishing
// Move or extinguish urgent fires, small fires or fires in that order.
// Wander status: move to refuges and unvisited buildings


package phoenix;

import java.net.*;
import java.util.*;
import yab.agent.*;
import yab.agent.object.*;

public class FireBrigadeAgent extends AbstractFireBrigadeAgent {
	//??? still no fix answer for this amount
	protected static final int REQUIRED_WATER = 1000;
	
	private HashSet unknownBuildings = new HashSet();
	//private HashSet unvisitedRefuges = new HashSet();
	
	private HashSet burningBuildings = new HashSet();
	private HashSet urgentBuildings = new HashSet(); //burning buildings who spread fires
	private HashSet buriedBurningBuildings = new HashSet();
	private HashSet firePoints = new HashSet();
	
	private HashSet buriedBuildings = new HashSet(); //the building which this firebrigade is buried for locate message
	private HashSet blockedRoads = new HashSet(); //for clear message
	
	//status, 0:initial 1:refill water
	private int status;
	protected static final int SEARCH=0, EXTINGUISH=1, REFILL=2, GOTARGETS=3;
	protected Building target;
	private Random random;
	private int hearMessages;
	private int prevDegree = 0;
	
	protected static final Property
		FIERYNESS_PRP    = Property.get("Building", "fieryness"),
		//BURNING_TIME_PRP = Property.get("Building", "burningTime"),
		ENTRANCE_PRP     = Property.get("Building", "entrance"),
		/*add for civilian */
		POSITION_PRP   = Property.get("MovingObject", "position");
	
	//protected static final int BURNING_TIME_OF_AN_EARLY_FIRE = 3;
	
	protected static final Condition IS_BURNING_CND = FIERYNESS_PRP.gte(1).and(FIERYNESS_PRP.lte(3));
	protected static final Condition IS_NOT_PUTOUT = FIERYNESS_PRP.lte(3);    
	//obsolete, only for undelayed fires
	//IS_EARLY_CND   = BURNING_TIME_PRP.lte(BURNING_TIME_OF_AN_EARLY_FIRE);
	protected final Condition isNearCnd = distancePrp.lte(EXTINGUISHABLE_DISTANCE);
		
	public FireBrigadeAgent(InetAddress address, int port) {
        	super(address, port);
	
		random = new Random(self().id);
	
		status = SEARCH;

		
		//set unknown buildings
		for (Iterator it=world.buildings.iterator(); it.hasNext(); ) {
			Building b=(Building)it.next();
			if (!(b instanceof Refuge)) 
				unknownBuildings.add(b);
		} //end for
		

	} //end FireBrigeade()
	
	protected MotionlessObject mlpos() { return self().motionlessPosition(); }
	
	//Main program for an agent's action	
	protected void act() throws ActionCommandException {
		//debug, move begins from cycle 4 and far fires from cycle 6 
		/*if (self().id==2406) {
		Building b= (Building)world.get(1844);
		Building b1= (Building)world.get(1878);
		System.out.println("time "+time()+" "+self()+" "+mlpos()+b.details());
		move(b1.entrance());
		}	
		*/
    
		hearMessages = 0;
		    
		checkSelf();
		
		//a method in Humanoid program. if an agent cannot move in two cycles
		//temp: move to nearby buildings
		if (checkTrafficJam()) {
			moveToBuilding();
		}
		
		supplyWater();
		
		//if the target is putout, move to a nearby fire. Otherwise extinguish it
		if (status == EXTINGUISH) {
			//System.out.println("Time "+time()+" "+self()+" target "+target+" fieryness "+target.fieryness()+" status "+status);
			//if target is putout or needn't to extinguish(be a fire center), find another fire 
			int[] count=surrounding(target);
			if ( target.isPutOut() || (count[1] >= 3 && count[0]<=1) ) {
				status = SEARCH;
				
				firePoints.remove(target);
				buriedBurningBuildings.remove(target);
				urgentBuildings.remove(target);
				burningBuildings.remove(target);
						
				tell("time "+time()+" putout "+target.id);
				//System.out.println("!!!Time "+time()+" "+self()+" putout "+target);
				
				extinguishNearbyFires(firePoints);
				extinguishNearbyFires(buriedBurningBuildings);
				extinguishNearbyFires(urgentBuildings);
				extinguishNearbyFires(burningBuildings);
			} else {
				extinguish(target);
			}					
		} //end if
	
		//target = findUrgentFire();
		

		move_extinguish(firePoints);
		move_extinguish(buriedBurningBuildings);
		move_extinguish(urgentBuildings);
			
	        Collection smallFires = FIERYNESS_PRP.eq(1).extract(burningBuildings);
		if (!smallFires.isEmpty())
               		move_extinguish(smallFires);
		
		if (!burningBuildings.isEmpty())	
			move_extinguish(burningBuildings);
		
		wander();
	
	} //end act()

	//report blockade roads or buried building
	//collect fire information	
	protected void checkSelf() throws ActionCommandException {
		if (self().hp() == 0)
			rest();
		
		if (mlpos() instanceof Building) {
			unknownBuildings.remove((Building)mlpos());
		}
	
		//??? trouble in communication before cycle 4
		if (time() <= Global.CycleAgentFreeze)
			return;
			
		//firebrigade is buried	
		if (self().buriedness() > 0) {
			if (!(mlpos() instanceof Building))
				System.out.println("WARNING: "+self()+" is not in a building but at "+mlpos());
			
			int degree = 5; //the highest degree because buired firebrigades must be rscued at first
			Building b = (Building)mlpos();
			if (!buriedBuildings.contains(b)) {
				buriedBuildings.add(b);
				tell("time " + time() + " locate "+b.id+" degree "+degree);
			}		
		
			rest();
		}
			
		//report block road or itself is buried	
		if (mlpos() instanceof Road) {
			int degree = 1;
			if (status == GOTARGETS) {
				degree = 2;
			}
			if (prevDegree > 0)
				degree = 5;
			
			Road r = (Road)mlpos();
			if ( r.passableLines() == 0  ) {
				if (!blockedRoads.contains(r)) {
					blockedRoads.add(r);
				//System.out.println(r+" block "+r.block()+" repairCost "+r.repairCost()+" tohead "+r.linesToHead() +
				//" totail "+r.linesToTail()+" passable "+r.passableLinesToHead()+" "+r.passableLinesToTail());
					tell("time " + time() + " clear " + r.id + " degree "+degree);
					if (degree == 5)
						prevDegree = 0;
					else
						prevDegree = degree;	
				}	
			}
		}
		
		
		//check: damage>0, hp>0, hp>damage*buriedness, and NOT in refuge
		for(Iterator it=world.humanoids.iterator(); it.hasNext(); ) {
			Humanoid h = (Humanoid)it.next();
		
			if (h.hp() > 0 && h.damage() >0 && h.time() == time()) {
				MotionlessObject ml = h.motionlessPosition();
				//tell the building
				if (ml instanceof Building && !(ml instanceof Refuge)) {
					//System.out.println("Time "+time()+" "+self()+" tell "+ml);
					tell("time "+time()+" locate "+ml.id);
					break;
				}
			}
		} // end for
					
		//update burning building burning information
		checkFires();
	
   		for(Iterator it=burningBuildings.iterator(); it.hasNext(); ) {
			Building b = (Building)it.next();
			if (b.isPutOut())
				it.remove();
		}
	
		for(Iterator it=urgentBuildings.iterator(); it.hasNext(); ) {
			Building b = (Building)it.next();
			if (b.isPutOut())
				it.remove();
		}
	
		for(Iterator it=buriedBurningBuildings.iterator(); it.hasNext(); ) {
			Building b = (Building)it.next();
			if (b.isPutOut())
				it.remove();
		}	
		
		if (self().damage()>0) {
			System.out.println("!!!Time "+time()+" "+self()+" buried "+self().buriedness()+" hp "+self().hp());
			move(world.refuges);
		}	
	} //end checkSelf()

	protected void supplyWater() throws ActionCommandException {
	
		if (self().waterQuantity() == 0) {
			status = REFILL;
			moveToRefuges();
		}
	
		if (status == REFILL) {
			if (self().waterQuantity() >= REQUIRED_WATER) {
				status = SEARCH;
			} else {
				moveToRefuges();
			}
		}		
	} //end supplyWater()

	protected void checkFires() {
		//temporarily use all buildings 
		for(Iterator it=world.buildings.iterator(); it.hasNext(); ) {
			Building b = (Building)it.next();

			if (b.time()==time() && b.isBurning()) {
			
				if (!urgentBuildings.contains(b)) {
					//check b whether is a urgent fire, count[0] unburned; [1]burning; [2]putout
					int[] count = surrounding(b);

					//few surroundings are burning. eg urgent fires
					if (count[1] <= 1 ) {
					//System.out.println("Time "+time()+" "+self()+" "+b+" surrounding "+count[0]+" "+count[1]+" "+count[2]);
						//System.out.println("Time "+time()+" "+self()+" "+b.details());
						urgentBuildings.add(b);
						//if too many fires, don't send more messages. 15 is a arbitrary number
						if (urgentBuildings.size() < 15) {
							String messages = "time "+time()+" extinguish "+b.id;
							tell(messages);
						}	
					} else {
						burningBuildings.add(b);
					}
				}	
			}
		} //end for
	
		//find firePoints
		firePoints.clear();
		for(Iterator it=urgentBuildings.iterator(); it.hasNext(); ) {
			Building b = (Building)it.next();
			int[] count = surrounding(b);
			//few surroundings are burning
			if (count[1] <= 1 ) {
				firePoints.add(b);
			}	
		}
		
		
		//find burning buildings with buried persons
		for(Iterator it=world.humanoids.iterator(); it.hasNext(); ) {
			Humanoid h = (Humanoid)it.next();
			if (h.hp()>0 && h.damage()>0 && h.time()==time()) {
				MotionlessObject ml = h.motionlessPosition();
				if (ml instanceof Building) {
					Building b= (Building)ml;
					if (b.isBurning()) {
						buriedBurningBuildings.add(b);
						String messages = "time "+time()+" extinguish "+b.id;
						tell(messages);
					}	
				
				}
			
			}
		
		} // end for humanoid
	} //end checkFires()

	//get the status of surrounding buildings within the maximum of 10m and the length of this building side.
	//count[0]:unburned, count[1]:burning, and count[2]:putout
	protected int[] surrounding(Building current) {
		//status: 0 unburned, 1 burning, 2 putout
		int[] count = new int[3];
	
		int length = (int)Math.sqrt(current.buildingAreaGround());
		length = Math.max(length, 10000);
	
		for(Iterator it=world.buildings.iterator(); it.hasNext(); ) {
			Building b = (Building)it.next();
			//get current information
			if (b != current && b.distance(current) < length) {
				if (b.isUnburned()) {
					count[0]++;
				} else if (b.isBurning()) {
					count[1]++;
				} else {
					count[2]++;
				}
			}		
		}
	
		return count;   
	}
	
	//Avoid traffic jam near fires. If an agent is in the node or a road with passble line less than 1,
	// it move to nearby palce of the target building, usually a fire
	protected void avoidFireTrafficJam(Building b) throws ActionCommandException {
		boolean bMove = false;
		if (mlpos() instanceof Building)
			return;
		
		if (mlpos() instanceof Node)
			bMove = true;
		
		if (mlpos() instanceof Road) {
			Road r = (Road)mlpos();
			if ( r.passableLines()<=1 )
				bMove = true;
		}

			/*	
			boolean nearPlatoons = false;
			for(Iterator it=world.fireBrigades.iterator(); it.hasNext(); ) {
				FireBrigade fb = (FireBrigade)it.next();
				//nearby fbs
				if (fb != self() && fb.time() == time()) {
					nearPlatoons = true;
					break;
				}	
			}
			
			for(Iterator it=world.ambulanceTeams.iterator(); it.hasNext(); ) {
				AmbulanceTeam at = (AmbulanceTeam)it.next();
				//nearby fbs
				if (at.time() == time()) {
					nearPlatoons = true;
					break;
				}	
			}
			
			if (!nearPlatoons)
				return;
			*/
			
		if (bMove) {		
			//move to nearby buildings in any case
			System.out.println("Time "+time()+" "+self()+" avoid traffic jam and move to buildings near "+b);			
			moveNearbyBuildings(b);
		
		}
	} //end avoidFireTrafficJam()
	  
	//there are fires in extinguish distance, extingush a fire with mimimum fieryness
	//if the agent is in a burning building, move to nearby buildings
	//if the agent cause the traffic jam, move to nearby buildings
	protected void extinguishNearbyFires(Collection fires) throws ActionCommandException {
	//Collection fires = IS_BURNING_CND.extract(burningBuildings);
		Collection nearFires = isNearCnd.extract(fires);
		if (!nearFires.isEmpty()) {
			status = EXTINGUISH;
			Building b = (Building)FIERYNESS_PRP.min(nearFires);
			target = b;
		
			//current in burning buildings
			if (mlpos() instanceof Building) {
				Building mlb = (Building)mlpos();
				if (mlb.isBurning())
					moveNearbyBuildings(target);
			}		
			
			//avoid traffic jam
			//temporary disbale for qualification
			//avoidFireTrafficJam(target);
		
			System.out.println("Time "+time()+" "+self()+" extinguish "+b.details());
			extinguish(target);
		}	
	}
	
	//find a urgent burning buildings with the minimum distance.
	//if the found fire is differnet with current target fire, set target to it.
	protected Building findUrgentFire() {
		status = SEARCH;
		Collection urgent = IS_NOT_PUTOUT.extract(urgentBuildings);
		if (!urgent.isEmpty()) {
			Building b = (Building)distancePrp.min(urgent);
		
			if ( b != target)
				System.out.println("Time "+time()+" "+self()+" aims "+b.details());
		
			return b;
		}
		
		return null;	
	}	

    protected void move_extinguish(Building b) throws ActionCommandException {
	if (isNearCnd.eval(b)) {
		status = EXTINGUISH;
		
		//current in burning buildings
		if (mlpos() instanceof Building) {
			Building mlb = (Building)mlpos();
			if (mlb.isBurning())
				moveNearbyBuildings(b);
		}
		
		avoidFireTrafficJam(b);
		
		System.out.println("Time "+time()+" "+self()+" extinguish "+b.details());	
		extinguish(b);
	} else {
		//get roads in 30m, EXTINGUISHABLE_DISTANCE
		/*fail to do so
		HashSet places = new HashSet(); 
		for(Iterator it=world.nodes.iterator(); it.hasNext(); ) {
			Node r = (Node)it.next();
			if ( b.distance(r) < EXTINGUISHABLE_DISTANCE )
				places.add(r);
		}		 
		move(places);
		*/
		status = GOTARGETS;
		//move(b.entrance());
		
		//move to all nearby nodes
		moveNearbyBuildings(b);
	}	
    }

       
	protected void move_extinguish(Collection fires) throws ActionCommandException {
		if (fires.isEmpty())
			return;
		
		if (time()<30)	
			System.out.println("Time "+time()+" "+self()+" "+(fires.size()<10 ? fires.toString():"Too many fires") );	
 		//if a fire is in extinguishable distance, extinguish it
		extinguishNearbyFires(fires);
	
		//move to fires
		status = GOTARGETS;
	
		//move(ENTRANCE_PRP.collect(fires));
		
		//move to a nearby unburned building around fires, then extinguish a fire
		moveNearbyBuildings(fires);
	} //end move_extinguish()

	//move the building nearby a fire then extinguish it    
	protected void moveNearbyBuildings(Building fire) throws ActionCommandException {
		moveNearbyBuildings(Collections.singleton(fire));
	}
	
	protected void moveNearbyBuildings(Collection fires) throws ActionCommandException {
		HashSet nearBuildings = new HashSet();
	
		//nearby unburned buildings around fires
		for(Iterator it=world.buildings.iterator(); it.hasNext(); ) {
			Building b = (Building)it.next();
			if (!b.isBurning() && !fires.contains(b)) {
				for(Iterator f=fires.iterator(); f.hasNext(); ) {
					Building fire=(Building)f.next();
					if (b.distance(fire) < EXTINGUISHABLE_DISTANCE ) {
						nearBuildings.add(b);
						break;
					}
				}
			}
		} //end for
	
		if (!nearBuildings.isEmpty())
			move(nearBuildings);
		else {
			System.out.println("Time "+time()+" "+self()+" no nearby building, then move to entrances of fires "+fires);
			//move to the entrance node??? modify later if it doesn't work
			move(ENTRANCE_PRP.collect(fires));
		}
	} //end moveNearbyBuildings()
    
	protected void moveToRefuges() throws ActionCommandException {
		if (mlpos() instanceof Refuge) {
			rest();
		} else {
	        	move(world.refuges);
		}
	} //end moveToRefuges()

	
	protected void wander() throws ActionCommandException {
		//wander among refuges and unvisited buildings
		status = SEARCH;
	

		if (!unknownBuildings.isEmpty())
			move(unknownBuildings);	
/*		
	if (!unknownBuildings.isEmpty()) {
		int index = random.nextInt(unknownBuildings.size());
		
		Iterator it = unknownBuildings.iterator();
		for (int i=0;  i<index-1;  i++)
      			it.next();
		
		Building b= (Building)it.next();
		System.out.println("time "+time()+self()+" move "+b+b.x()+" "+b.y());	
		move(b);
	}	
*/

		System.out.println("Time "+time()+" "+self()+" NO PLACE TO GO");	
	} //end wander()

	//main communcation method
	//message from the fire station: time T extinguish/putout id1 ... idn
	//discard message: time T locate id1 ... idn
	protected void hear(RealObject sender, String message) {
		if (sender instanceof FireBrigade)
			return;
	
		if (hearMessages >= hearingLimit())
			return;
	
		if (sender instanceof FireStation) {
			//System.out.println("time "+time()+" "+self()+" "+sender+message);
			hearMessages++;
				
			StringTokenizer st = new StringTokenizer(message);
			st.nextToken(); //time
        		st.nextToken();
		
			String command = st.nextToken();

			//st.nextToken(); //fieryness
		
			if (command.equals("extinguish")) {
				while(st.hasMoreTokens()) {
					RealObject target = world.get(Integer.parseInt(st.nextToken()));
					if (target instanceof Building) {
						urgentBuildings.add(target);
						unknownBuildings.remove(target);
					}	
				}	
			} else if (command.equals("putout")) {
				while(st.hasMoreTokens()) {
					RealObject target = world.get(Integer.parseInt(st.nextToken()));
					buriedBurningBuildings.remove(target);
					urgentBuildings.remove(target);
					burningBuildings.remove(target);
					unknownBuildings.remove(target);
				}	
			}
		} //end if	
		
	} // end hear()
    
} //end class
