// Initial Developer: 2002 Takeshi Morimoto <morimoto@takopen.cs.uec.ac.jp> 
// Author: Zijian Ren
// Version :1.10
// Date: Sep 16, 2003
// Comments: Add a collection of messages
//		Add act() module

// clear message from firebrigade: time X id Y degree Z

package phoenix;

import java.net.*;
import yab.agent.*;
import yab.agent.object.*;
import java.util.*;

public class FireStationAgent extends AbstractFireStationAgent {
	
	//Collection of clear roads. Data Stucture: id|messagebase
	//where messageBase includes sentflag, receiveTime, sendTime, degree, and comment
	
	private HashMap blockadeRoads = new HashMap();
	private HashMap buriedBuildings = new HashMap();
	
	//Insesction of Burning and Putout Set is empty
	private HashMap burningBuildings = new HashMap();
	private HashMap putoutBuildings = new HashMap();
	
	
		
    public FireStationAgent(InetAddress address, int port) {
        super(address, port);
    }

    protected void act() throws ActionCommandException {
    	//System.out.println("time "+time()+" "+self()+blockadeRoads);
	String body = "";
					
	body = Global.tellMessages(blockadeRoads, time());
	if (!body.equals("")) {
		String message = "time "+time()+" clear "+body;
		tell(message);
		
		boolean verbose = false;	
		if (verbose)
			System.out.println("Time "+time()+" "+self()+" Message: "+message);
	}		
	
	//tellsMessages is an obselote method with simple data structure: id|times
	tellMessages(burningBuildings, "extinguish", true);
	tellMessages(putoutBuildings, "putout", true);
	
	body = Global.tellMessages(buriedBuildings, time());
	if (!body.equals("")) {
		String message = "time "+time()+" locate "+body;
		tell(message);
		
		boolean verbose = true;	
		if (verbose)
			System.out.println("Time "+time()+" "+self()+" Message: "+message);
	}		
	
		
    }

    	//old tell messages
        protected void tellMessages(Map messages, String command, boolean verbose) throws ActionCommandException {
		String message = "time "+ time() +" "+command;
		String ids = "";
		
		for(Iterator it=messages.entrySet().iterator(); it.hasNext(); ) {
			Map.Entry e = (Map.Entry)it.next();
			if ( !((String)e.getValue()).equals("0") ) {
				ids += (" " + (String)e.getKey());
				//reset this message time stamp to 0
				e.setValue("0");
			}
		}
		
		if (!ids.equals("")) {
			tell(message+ids);
			
			if (verbose)
				System.out.println("Time "+time()+" "+self()+" Message: "+message+ids);
		}		
	}


	
    protected void hear(RealObject sender, String message) {
	//System.out.println("Time "+time()+" receiver "+self()+" sender "+sender+" "+message);
        if (sender instanceof PoliceOffice || sender instanceof AmbulanceCenter) {
		StringTokenizer st = new StringTokenizer(message);
		st.nextToken();
        	String time = st.nextToken();
		String command = st.nextToken();
		
		while (st.hasMoreTokens()) {
			String id = st.nextToken();
		
			if (command.equals("extinguish")) {
				if (!burningBuildings.containsKey(id)) {
					burningBuildings.put(id, time);
				}	
			} else if (command.equals("clear")) {
				//reset all messages
				//blockadeRoads.put(id, "0");
			} else if (command.equals("locate")) {
				//reset all messages
				//buriedBuildings.put(id, "0");
			}	
		}
	} //end if
	
	if (sender instanceof FireBrigade) {
		//format: Time T Command id (degree D)
		//System.out.println("time "+time()+" "+self()+message);
		StringTokenizer st = new StringTokenizer(message);
		
		st.nextToken(); //String "Time"
        	String time = st.nextToken();
		String command = st.nextToken();
		
		String id = st.nextToken();
		
		if (command.equals("extinguish")) {
			//get a fire information. Check whether the fire is in Burning and Putout sets
			if (!putoutBuildings.containsKey(id) && !burningBuildings.containsKey(id))
				burningBuildings.put(id, time);
		} else if (command.equals("putout")) {
			//get a putout information. remove it from Burning Set
			if (!putoutBuildings.containsKey(id)) {
				putoutBuildings.put(id, time);
				burningBuildings.remove(id);
			}	
		} else if (command.equals("clear")) {
			//System.out.println(sender.id+message);
			int degree = 1; //default degree, the lowest one
			
			if (st.hasMoreTokens()) {
				st.nextToken(); //String "degree"
				degree = Integer.parseInt(st.nextToken());
			}
			
			if (!blockadeRoads.containsKey(id)) {
				MessageBase mb = new MessageBase(time(), degree);
				blockadeRoads.put(id, mb);
				
			} else {
				MessageBase mb = (MessageBase)blockadeRoads.get(id);
				if (degree > mb.degree) {
					mb.setValue(time(), degree);
				}	
			}	
				
			//if (!blockadeRoads.containsKey(id))
			//	blockadeRoads.put(id, time);
				
		} else if (command.equals("locate")) {
			int degree = 1; //default degree, the lowest one
			
			if (st.hasMoreTokens()) {
				st.nextToken(); //String "degree"
				degree = Integer.parseInt(st.nextToken());
			}
			
			if (!buriedBuildings.containsKey(id)) {
				MessageBase mb = new MessageBase(time(), degree);
				buriedBuildings.put(id, mb);
				
			} else {
				MessageBase mb = (MessageBase)buriedBuildings.get(id);
				if (degree > mb.degree) {
					mb.setValue(time(), degree);
				}	
			}
			//if (!buriedBuildings.containsKey(id))
			//	buriedBuildings.put(id, time);
		}

	
	} //end if sender
	
    }
    
  
}
