// Police Force agent: its purpose is to clear blockades to help other agents pass the roads.
// Author: Zijian Ren
// Version :1.20
// Date: Sep 23, 2003

package phoenix;

import java.net.*;
import java.util.*;
import yab.agent.*;
import yab.agent.object.*;

public class PoliceForceAgent extends AbstractPoliceForceAgent {

	//blocked degree from 5 to 1. unblocked is 0
    private HashSet blockadeRoads[] = new HashSet[6];
    private HashSet unknownRoads = new HashSet();
        
    //fire brigades is blocked by urgent roads
    //private HashSet blockadeRoads[2] = new HashSet();
    
    private HashSet buriedBuildings = new HashSet();
    private HashSet burningBuildings = new HashSet();

    private int status;
    private int hearMessages;
    private Random rand;
    protected static final int SEARCH=0, CLEAR=1, GOTARGETS=2;     
   
    protected static final Condition HAS_NOT_BEEN_SEEN_CND
        = Property.get("RealObject", "hasBeenSeen").equal(Boolean.FALSE);

	protected MotionlessObject mlpos() { return self().motionlessPosition(); }

    public PoliceForceAgent(InetAddress address, int port) {
        super(address, port);
		
	for(int i=0; i<6 ;i++) {
		blockadeRoads[i] = new HashSet();
	}	

		
	for(Iterator it=world.roads.iterator(); it.hasNext(); ) {
		unknownRoads.add((Road)it.next());
	}
	
	status = SEARCH;
	
	rand = new Random();
   }
	
    protected void act() throws ActionCommandException {
   /*        List uncheckedRoads = HAS_NOT_BEEN_SEEN_CND.extract(world.roads);
	   for(Iterator it=uncheckedRoads.iterator(); it.hasNext(); ) {
	   	Road r = (Road)it.next();
		if (r.block() > 0 || r.repairCost() > 0)
	   	System.out.print(" "+r+" x "+r.x()+" block "+r.block()+" repairCost "+r.repairCost());
	}	
	List uncheckedBuildings = HAS_NOT_BEEN_SEEN_CND.extract(world.buildings);
	   for(Iterator it=uncheckedBuildings.iterator(); it.hasNext(); ) {
	   	Building r = (Building)it.next();
		if (r.fieryness() > 0 || r.buildingCode() ==  0)
	   	System.out.print(" "+r+" x "+r.x()+" fieryness "+r.fieryness()+" buildingCode "+r.buildingCode());
	}	
*/
	hearMessages = 0;

	checkBlockades();
	
	checkSelf();
	
        clearHere();
	   
        moveToBlockades();
    
        wander();
    }

    protected void checkBlockades() {
        //check recent seen roads
	for(Iterator it = world.roads.iterator(); it.hasNext(); ) {
		Road r = (Road)it.next();
		if (r.time() == time()) {
			
			//there are three conditions
			//unknown before(degree=-1), unblocked (degree=0) and blocked (degree>=1)
			int degree = getDegree(r);
			if (degree == -1) {
				if (r.passableLines()==0) {
					blockadeRoads[1].add(r);
				} else {
					blockadeRoads[0].add(r);
				}
				//remove from unknown roads
				unknownRoads.remove(r);
			} else if (degree == 0) {
				if (r.passableLines() == 0) {
					//road is blocked
					blockadeRoads[1].add(r);
					blockadeRoads[0].remove(r);
				}
			} else {
				if (r.passableLines() > 0) {
					//road is unblocked
					blockadeRoads[0].add(r);
					blockadeRoads[degree].remove(r);
				}	
			}
			
		}	
	} //end for
	

/*	
	for (Iterator it = blockadeRoads[2].iterator();  it.hasNext();  ) {
            Road road = (Road) it.next();
            if (road.hasBeenSeen()  &&  road.passableLines() > 0)
                it.remove();
        }

	for (Iterator it = blockadeRoads[1].iterator();  it.hasNext();  ) {
            Road road = (Road) it.next();
            if (road.hasBeenSeen()  &&  road.passableLines() > 0)
                it.remove();
        }
 */       
	
    }

    private int getDegree(Road r) {
    	for(int i=5; i>=0; i--) {
		if (blockadeRoads[i].contains(r))
			return i;
	}
	
	return -1; //unvisited road		
    }
    
    	
    protected void checkSelf() throws ActionCommandException {
    	//??? at cycle 3, tell doesn't work
    	if (time() > Global.CycleAgentFreeze)
		reportBuildingsAndFires();
	
	if (self().hp() == 0) 
		rest();	    
    
	//policeforce is buried	
	if (self().buriedness() > 0) {
		if (!(mlpos() instanceof Building))
			System.out.println("WARNING: "+self()+" is not in a building but at "+mlpos());
			
		int degree = 4; //the highest degree because buired firebrigades must be rscued at first
		Building b = (Building)mlpos();
		if (!buriedBuildings.contains(b)) {
			buriedBuildings.add(b);
			tell("time " + time() + " locate "+b.id+" degree "+degree);
		}		
		
		rest();
	}
	
		
	if (self().damage()>0) {
		System.out.println("!!!Time "+time()+" "+self()+" buried "+self().buriedness()+" hp "+self().hp());
		move(world.refuges);
	}	
    }
    
    
    protected void reportBuildingsAndFires() throws ActionCommandException {
    	//only after some cycles
	if (time()>35) {
	for(Iterator it=world.civilians.iterator(); it.hasNext(); ) {
		Humanoid h = (Humanoid)it.next();
		
		if (h.hp() > 0 && h.damage() >0 && h.time() == time()) {
			MotionlessObject ml = h.motionlessPosition();
			//tell the building
			if (!(ml instanceof Refuge)) {
				if (ml instanceof Building && !buriedBuildings.contains(ml)) {
					//System.out.println("Time "+time()+" "+self()+" tell "+ml);
					buriedBuildings.add(ml);
					tell("time "+time()+" locate "+ml.id);
					break;
				}	
			}
		}
	
	} //end for
	}
	
	if (time()<20) {
	for(Iterator it=world.buildings.iterator(); it.hasNext(); ) {
		Building b = (Building)it.next();
		
		if (b.isBurning() && !burningBuildings.contains(b)) {
			//System.out.println("Time "+time()+" "+self()+" tell: extinguish "+b.id);
			burningBuildings.add(b);
			tell("time "+time()+" extinguish "+b.id);
			break;
		}
	} //end for
 	}
    }

    protected void clearHere() throws ActionCommandException {
	if (mlpos() instanceof Road) {
		Road r = (Road)mlpos();
                if ( r.passableLines() == 0) {
			status = CLEAR;
			//for small roads, only need one PF
			if ((r.linesToHead() + r.linesToTail()) < 4) {
				//System.out.println("Time "+time()+" PoliceForce "+self().id+": finish road "+r.id);
				tell("time "+time()+" remove "+r.id);
			}	
			clear(r);
		} else if (status==CLEAR) {
			status = SEARCH;
			//report only the road with degree =5
			//if (blockadeRoads[5].contains(r)) {

			//}	
		}	
			
	}

    }

    protected void moveToBlockades() throws ActionCommandException {
    	//Move to serious blocakade roads
    	for(int i=5; i>=2; i--) {
        	if (!blockadeRoads[i].isEmpty()) {
        		move_N(blockadeRoads[i]);
		}
	}	

	//Move to nearby blockades
	for(Iterator it = world.roads.iterator(); it.hasNext(); ) {
		Road r = (Road)it.next();
		if (r.time()==time() && r.passableLines()==0 && r!=mlpos())
			move(r);
	}		
		
	//special treatment for normal blockade (degree = 1)
	if (!blockadeRoads[1].isEmpty()) {
		HashSet wideRoads = new HashSet();
		for(Iterator it=blockadeRoads[1].iterator(); it.hasNext(); ) {
			Road r= (Road)it.next();
			if ( (r.linesToHead() + r.linesToTail()) > 2)
				wideRoads.add(r);
		}
		
		if (!wideRoads.isEmpty())
			move_N(wideRoads);
		else
			move_N(blockadeRoads[1]);
	}
	
		    
    }


    protected void wander() throws ActionCommandException {
	if (!unknownRoads.isEmpty())
		move(unknownRoads);
	
	System.out.println("Time "+time()+" "+self()+" NO PLACE TO GO");    
    }

    protected void hear(RealObject sender, String message) {
    
    //System.out.println("Time "+time()+" receiver "+self()+" sender "+sender+" "+message); 
	//if adding messages from PolicForce, the message from PolicOffice may lost.    
        if (sender instanceof PoliceOffice) {
	
		if (hearMessages >= hearingLimit())
			return;
	
		hearMessages++;
		
        		StringTokenizer st = new StringTokenizer(message);
			st.nextToken(); //"Time"
			st.nextToken(); //value of time
        		String command = st.nextToken();
	
			if (command.equals("clear")) {
				int degree = 1;
				//clear degree 5 x1 x2 ... degree 1 y1 y2  
				while(st.hasMoreTokens()) {
					String s = st.nextToken();
					if (s.equals("degree")) {
						degree = Integer.parseInt(st.nextToken());
					} else {
						RealObject target = world.get(Integer.parseInt(s));
						int currentDegree = getDegree((Road)target);
						
						//random pick 3 PF to receive degree >=2
						int x=rand.nextInt(world.policeForces.size());
						
						int wideness = ((Road)target).linesToHead()+((Road)target).linesToTail();
						if ( (degree>=2 && x<=wideness) || degree==1 ) {
							//Road has no degree
							if (currentDegree == -1) {
								blockadeRoads[degree].add(target);
								unknownRoads.remove(target);
							} else if (degree > currentDegree) {
								//only modify when the reported degree is more serious
								blockadeRoads[degree].add(target);
								blockadeRoads[currentDegree].remove(target);
							}
						}	
					}
				} //end while
				
			} else if (command.equals("remove")) {
				while (st.hasMoreTokens()) {
					String s = st.nextToken();
					RealObject target = world.get(Integer.parseInt(s));
					for(int i=5; i>=1; i--) {
        					blockadeRoads[i].remove(target);
					}
						
					blockadeRoads[0].add(target);
					unknownRoads.remove(target);
				}		
			}

	} // end if sender	
    } // end hear()
}
